import unittest
import numpy as np
import pandas as pd
import json
import priceprop.propagator as prop

# some basic checks, not meant to replace manual testing with real data!

class TestIntegrate(unittest.TestCase):
    def test(self):
        np.testing.assert_almost_equal(prop.integrate([1,2,3]), [0,1,3])
        
class TestSmoothTailRBF(unittest.TestCase):
    def test(self):
        x = np.arange(1,100, dtype=float)**-1
        s = prop.smooth_tail_rbf(x)
        np.testing.assert_allclose(s, x, rtol=.01, atol=.01)

class TestPropagate(unittest.TestCase):
    def test_pulse(self):
        np.testing.assert_almost_equal(
            prop.propagate([1,0,0], [1,.5,0]),
            [1,.5,0]
        )
        
class TestResponse(unittest.TestCase):
    
    def test_lsr(self):
        # test simple response
        args = ([-1,0,1,0], [0,1,0,0])
        l,s,r = prop.response(*args)
        np.testing.assert_almost_equal(
            l, range(-2,3)
        )
        np.testing.assert_almost_equal(
            s, [0, -.25, 0, .25, 0]
        )
        np.testing.assert_almost_equal(
            r, [.25,.25,0,0,.25]
        )
        # test ret argument
        np.testing.assert_almost_equal(
            l, prop.response(*args, ret='l')
        )
        np.testing.assert_almost_equal(
            s, prop.response(*args, ret='s')
        )
        np.testing.assert_almost_equal(
            r, prop.response(*args, ret='r')
        )
        
    def test_grouped(seld):
        df = pd.DataFrame({
            'r':    [-2, 0, 1, 0,-4, 0, 1, 0],
            's':    [ 0, 1, 0, 0, 0, 1, 0, 0], 
            'date': [ 0, 0, 0, 0, 1, 1, 1, 1]
        })
        l, s, r = prop.response_grouped_df(df, ['r','s'])
        np.testing.assert_almost_equal(
            l, range(-2,3)
        )
        np.testing.assert_almost_equal(
            s, [ 0., -0.75, 0., 0.25, 0.]
        )
        np.testing.assert_almost_equal(
            r, [.75,.75,0,0,.25]
        )
        
        
class TestTIM1(unittest.TestCase):    
    def test_tim1(self):
        np.testing.assert_almost_equal(
            prop.tim1([1,0,0,0], [1,.5,0,0]), 
            [1,.5,0,0]
        )

    def test_calibrate_tim1(self):
        np.testing.assert_almost_equal(
            prop.calibrate_tim1([.25,0,0,0], [0,0,0,.25,.125,0,0], maxlag=4),
            [1,.5,0,0]
        )
        
class TestTIM2(unittest.TestCase):
    def test_tim2(self):
        np.testing.assert_almost_equal(
            prop.tim2([1,-1], np.array([1,0,], dtype=bool),[1,0,],[.1,0,]),
            [.1,-1]
        )
    def test_calibrate_tim2(self):
        gn, gc = prop.calibrate_tim2(
            [1,0],[1,0],[0,0],[0,0],[0,.1,0],[0,1,0], maxlag=2
        )
        np.testing.assert_almost_equal(gn, [.1,0])
        np.testing.assert_almost_equal(gc, [1,0])
        
class TestHDIM2(unittest.TestCase):
    def test_hdim2(self):
        np.testing.assert_almost_equal(
            prop.hdim2([1,-1], np.array([1,0,], dtype=bool),[1,0,],[.1,0,]),
            [.1,0]
        )
        
    def test_calibrate_hdim2(self):
        # It is somehow difficult to find tiny example like above where the 
        # three-point correlation matrix isn't singular!
        with open('tests/hdim2_test.json', 'r') as f:
            lpars = {
                k: np.array(v) if hasattr(v, '__len__') 
                else v 
                for k,v in json.load(f).iteritems()
            }
        kn_est, kc_est = prop.calibrate_hdim2(
            lpars['Cnnc'], 
            lpars['Cccc'], 
            lpars['Ccnc'], 
            lpars['Sn'], 
            lpars['Sc'], 
            maxlag=lpars['maxlag']
        )
        np.testing.assert_allclose(lpars['kn'], kn_est, rtol=.01, atol=.01)
        ## The required file can be saved as follows:
        #savepars = {
        #    'Cnnc': Cnnc[:maxlag,:maxlag], # an estimation
        #    'Cccc': Cccc[:maxlag,:maxlag],
        #    'Ccnc': Ccnc[:maxlag,:maxlag],
        #    'Sn':   Sn[:2*maxlag],
        #    'Sc':   Sc[:2*maxlag],
        #    'maxlag': maxlag,
        #    'kn':   kn, # the real kernel
        #    'kc':   kc
        #}
        #with open('hdim2_test.json', 'w') as f:
        #    json.dump(
        #        {
        #            k: v.tolist() if hasattr(v, '__len__') 
        #            else v 
        #            for k,v in savepars.iteritems()
        #        },
        #        f
        #    )


import numpy as np
from scipy.linalg import solve_toeplitz, solve
from scipy.signal import fftconvolve
from scipy.interpolate import Rbf
from scorr import xcorr, xcorr_grouped_df, xcorrshift, fftcrop, corr_mat


# Helpers
# =====================================================================

def integrate(x):
    "Return lag 1 sum, i.e. price from return, or an integrated kernel."
    return np.concatenate([[0], np.cumsum(x[:-1])])
    
    
def smooth_tail_rbf(k, l0=3, tau=5, smooth=1, epsilon=1):
    """Smooth tail of array k with radial basis functions"""
    # interpolate in log-lags
    l = np.log(np.arange(l0,len(k)))
    # estimate functions
    krbf = Rbf(
        l, k[l0:], function='multiquadric', smooth=smooth, epsilon=epsilon
    )
    # weights to blend with original for short lags
    w = np.exp(-np.arange(1,len(k)-l0+1)/ float(tau))
    # interpolate
    knew     = np.empty_like(k)
    knew[:l0]  = k[:l0]
    knew[l0:] = krbf(l) * (1-w) + k[l0:] * w
    #done
    return knew
    
def propagate(s, G, sfunc=np.sign):
    """Simulate propagator model from signs and one kernel.
    Equivalent to tim1, one of the kernels in tim2 or hdim2.
    """
    steps = len(s)
    s  = sfunc(s[:len(s)])
    p = fftconvolve(s, G)[:steps]
    return p

# Responses
# =====================================================================

def _return_response(ret, x, maxlag):
    """Helper for response and response_grouped_df."""
    # return what?
    ret = ret.lower()
    res = []
    for i in ret:
        if i   == 'l':
            # lags
            res.append(np.arange(-maxlag,maxlag+1))
        elif i == 's':
            res.append(
                # differential response
                np.concatenate([x[-maxlag:], x[:maxlag+1]])
            )
        elif i == 'r':    
            res.append(
            # bare response === cumulated differential response
                np.concatenate([
                    -np.cumsum(x[:-maxlag-1:-1])[::-1], 
                    [0], 
                    np.cumsum(x[:maxlag])
                ])
            )
    if len(res) > 1:
        return tuple(res)
    else:
        return res[0]

def response(r, s, maxlag=10**4, ret='lsr', subtract_mean=False):
    """Return lag, differential response S, response R.
    
    Note that this commonly used price response is a simple cross correlation 
    and NOT equivalent to the linear response in systems analysis.
    
    Parameters:
    ===========
    
    r: array-like
        Returns
    s: array-like
        Order signs
    maxlag: int
        Longest lag to calculate
    ret: str
        can include 'l' to return lags, 'r' to return response, and
        's' to return differential response (in specified order).
    subtract_mean: bool
        Subtract means first. Default: False (signal means already zero)
    """
    maxlag = min(maxlag, len(r) - 2)
    s  = s[:len(r)]
    # diff. resp.
    # xcorr == S(0, 1, ..., maxlag, -maxlag, ... -1)
    x = xcorr(r, s, norm='cov', subtract_mean=subtract_mean)
    return _return_response(ret, x, maxlag)

def response_grouped_df(
        df, cols, nfft='pad', ret='lsr', subtract_mean=False, **kwargs
    ):
    """Return lag, differential response S, response R calculated daily.
    
    Note that this commonly used price response is a simple cross correlation 
    and NOT equivalent to the linear response in systems analysis.
    
    Parameters
    ==========
    
    df: pandas.DataFrame
        Dataframe containing order signs and returns
    cols: tuple
        The columns of interest
    nfft:
        Length of the fft segments
    ret: str
        What to return ('l': lags, 'r': response, 's': incremental response).
    subtract_mean: bool
        Subtract means first. Default: False (signal means already zero)
    
    See also response, spectral.xcorr_grouped_df for more explanations
    """
    # diff. resp.
    x = xcorr_grouped_df(
        df, 
        cols,
        by            = 'date', 
        nfft          = nfft, 
        funcs         = (lambda x: x, lambda x: x), 
        subtract_mean = subtract_mean,
        norm          = 'cov',
        return_df     = False,
        **kwargs
    )[0]
    # lag 1 -> element 0, lag 0 -> element -1, ...
    #x = x['xcorr'].values[x.index.values-1]
    maxlag = len(x) / 2
    return _return_response(ret, x, maxlag)
   
# Analytical power-laws
# =====================================================================

def beta_from_gamma(gamma):
    """Return exponent beta for the (integrated) propagator decay 
        G(lag) = lag**-beta 
    that compensates a sign-autocorrelation 
        C(lag) = lag**-gamma.
    """
    return (1-gamma)/2.
    
def G_pow(steps, beta):
    """Return power-law Propagator kernel G(l). l=0...steps"""
    G = np.arange(1,steps)**-beta#+1
    G = np.r_[0, G]
    return G
    
def k_pow(steps, beta):
    """Return increment of power-law propagator kernel g. l=0...steps"""
    return np.diff(G_pow(steps, beta))
    
# TIM1 specific 
# =====================================================================

def calibrate_tim1(c, Sl, maxlag=10**4):
    """Return empirical estimate TIM1 kernel
    
    Parameters:
    ===========
    
    c: array-like
        Cross-correlation (covariance).
    Sl: array-like
        Price-response. If the response is differential, so is the returned
        kernel.
    maxlag: int
        length of the kernel.
    See also: integrate, g2_empirical, tim1
 
    """
    lS = int(len(Sl) / 2)
    g = solve_toeplitz(c[:maxlag], Sl[lS:lS+maxlag])
    return g

def tim1(s, G, sfunc=np.sign):
    """Simulate Transient Impact Model 1, return price or return.
    
    Result is the price p when the bare responses G is passed
    and the 1 step ahead return p(t+1)-p(t) for the differential kernel 
    g, where G == numpy.cumsum(g).
    
    Parameters:
    ===========
    
    s: array-like
        Order signs
    G: array-like
        Kernel
    
    See also: calibrate_tim1, integrate, tim2, hdim2.
    """
    return propagate(s, G, sfunc=sfunc)

# TIM2 specific
# =====================================================================

def calibrate_tim2(
        nncorr, cccorr, cncorr, nccorr, Sln, Slc, maxlag=2**10
    ):
    """
    Return empirical estimate for both kernels of the TIM2.
    (Transient Impact Model with two propagators)
    
    Parameters:
    ===========
    
    nncorr: array-like
        Cross-covariance between non-price-changing (n-) orders.
    cccorr: array-like
        Cross-covariance between price-changing (c-) orders.
    cncorr: array-like
        Cross-covariance between c- and n-orders
    nccorr: array-like
        Cross-covariance between n- and c-orders.
    Sln: array-like
        (incremental) price response for n-orders
    Slc: array-like
        (incremental) price response for c-orders
    maxlag: int
        Length of the kernels.
    
    See also: calibrate_tim1, calibrate_hdim2
    """
    # incremental response
    lSn = int(len(Sln) / 2)
    lSc = int(len(Slc) / 2)
    S = np.concatenate([Sln[lSn:lSn+maxlag], Slc[lSc:lSc+maxlag]])
    
    # covariance matrix
    mat_fn = lambda x: corr_mat(x, maxlag=maxlag)
    C = np.bmat([
        [mat_fn(nncorr), mat_fn(cncorr)], 
        [mat_fn(nccorr), mat_fn(cccorr)]
    ])
        
    # solve
    g = solve(C, S)
    gn = g[:maxlag]
    gc = g[maxlag:]
            
    return gn, gc

def tim2(s, c, G_n, G_c, sfunc=np.sign):
    """Simulate Transient Impact Model 2
    
    Returns prices when integrated kernels are passed as arguments
    or returns for differential kernels.
    
    Parameters:
    ===========
    s: array
        Trade signs
    c: array
        Trade labels (1 = change; 0 = no change)
    G_n: array
        Kernel for non-price-changing trades
    G_c: array
        Kernel for price-changing trades
    sfunc: function [optional]
        Function to apply to signs. Default: np.sign.
        
    See also: calibrate_tim2, tim1, hdim2.
    """
    assert c.dtype == bool, "c must be a boolean indicator!"
    return propagate(s * c, G_c) + propagate(s * (~c), G_n)

    
# HDIM2 specific
# =====================================================================

def calibrate_hdim2(
        Cnnc, Cccc, Ccnc, Sln, Slc,
        maxlag=None, force_lag_zero=True
    ):
    """Return empirical estimate for both kernels of the HDIM2.
    (History Dependent Impact Model with two propagators).
    
    Requres three-point correlation matrices between the signs of one 
    non-lagged and two differently lagged orders.
    We distinguish between price-changing (p-) and non-price-changing (n-)
    orders. The argument names corresponds to the argument order in 
    spectral.x3corr.
    
    Parameters:
    ===========
    Cnnc: 2d-array-like
        Cross-covariance matrix for n-, n-, c- orders.
    Cccc: 2d-array-like
        Cross-covariance matrix for c-, c-, c- orders.
    Ccnc: 2d-array-like
        Cross-covariance matrix for c-, n-, c- orders.
    Sln: array-like
        (incremental) lagged price response for n-orders
    Slc: array-like
        (incremental) lagged price response for c-orders
    maxlag: int
        Length of the kernels.
        
    See also: hdim2,
    """
    maxlag = maxlag or min(len(Cccc), len(Sln))/2
    
    # incremental response
    lSn = int(len(Sln) / 2)
    lSc = int(len(Slc) / 2)
    S = np.concatenate([
        Sln[lSn:lSn+maxlag], 
        Slc[lSc:lSc+maxlag]
    ])
    
    # covariance matrix
    Cncc = Ccnc.T
    C = np.bmat([
        [Cnnc[:maxlag,:maxlag], Ccnc[:maxlag,:maxlag]], 
        [Cncc[:maxlag,:maxlag], Cccc[:maxlag,:maxlag]]
    ])
    
    if force_lag_zero:
        C[0,0] = 1
        C[0,1:] = 0
    
    # solve
    g = solve(C, S)
    gn = g[:maxlag]
    gc = g[maxlag:]
    
    return gn, gc
    
def hdim2(s, c, k_n, k_c, sfunc=np.sign):
    """Simulate History Dependent Impact Model 2, return return.
    
    Parameters:
    ===========
    s: array
        Trade signs
    c: array
        Trade labels (1 = change; 0 = no change)
    k_n: array
        Differential kernel for non-price-changing trades
    k_c: array
        Differential kernel for price-changing trades
    sfunc: function [optional]
        Function to apply to signs. Default: np.sign.
        
    See also: calibrate_hdim2, tim2, tim1.
    """
    assert c.dtype == bool, "c must be a boolean indicator!"
    return c * (propagate(s * c, k_c) + propagate(s * (~c), k_n))



from setuptools import setup, find_packages

setup(
    name='priceprop',
    version='1.0.0',
    description=(
        'Calibrate and simulate linear propagator models for the price '
        'impact of an extrinsic order flow.'
    ),
    long_description="""
        The models and methods are explained in:
	
            Patzelt, F. and Bouchaud, J-P.:
            Nonlinear price impact from linear models
            Journal of Statistical Mechanics (2017, in print)
            Preprint at arXiv:1706.04163
    """,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.7',
        #'Programming Language :: Python :: 3',
        #'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    keywords=[
        'propagator', 'price impact', 'intra day', 'order flow', 
        'price return', 'Transient Impact Model', 'TIM', 
        'History Dependent Impact Model', 'HDIM'
    ],
    url='http://github.com/felixpatzelt/priceprop',
    download_url=(
      'https://github.com/felixpatzelt/priceprop/archive/1.0.0.gz'
    ),
    author='Felix Patzelt',
    author_email='felix@neuro.uni-bremen.de',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'scipy',
        'pandas',
        'scorr'
    ],
    include_package_data=True,
    zip_safe=False,
    test_suite='tests',
)


# Helpers to perform batch analysis of various propagator models.
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import scorr
import propagator as prop

# Helpers
# ============================================================================

def complete_data_columns(
        tt,
        split_dates=True
    ):
    """Fill in columns necessary for models"""
    
    if not 'sc' in tt:
        tt['sc'] = tt['sign'] * tt['change']
    if not 'sn' in tt:
        tt['sn'] = tt['sign'] * (~tt['change'])
        
    if split_dates and not 'sample' in tt:
        # two sample groups based on date
        dt = np.zeros(len(tt), dtype=int)
        dt[1:] = (
            np.cumsum(
                np.array([t.days for t in np.diff(tt['date'])], dtype=int)
            ) % 2
        )
        tt['sample'] = dt


def get_trade_split(tt, split_by):
    "Get dict of sample key-mask pairs for splitting trades into groups."
    if split_by:
            samples = tt[split_by].unique()
            masks   = {m: (tt[split_by] == m) for m in samples}
    else:
        masks   = {'all': slice(None)}
        
    return masks
    
def shift(x, n, val=np.nan):
    """Shift array, pad with fixed value.
    
    Example: Convert r_1 = p(t+1) - p(t) to causal return p(t) - p(t-1)
             without losing timesteps with pad(r_1, 1).
    """
    
    if n == 0:
        return x
    else:
        res = np.empty_like(x)
        if n > 0:
            res[:n] = val
            res[n:] = x[:-n]
        else:
            res[n:] = val
            res[:n] = x[n:]
        return res
    
# Analyse Trades
# ============================================================================

def calibrate_models(
        tt, # true trades df
        nfft='pad', 
        group=False,
        models = ['cs','tim1','tim2','hdim2','hdim2_x2']
    ):
    """Return dict with correlations, kernels, and responses.
    Calculate sign & price change correlations, response functions, 
    and fitted propagator kernels for trades in DataFrame.
    """
    
    # store results
    res = {}
    
    # intra day mask
    mask0 = ~tt['r1'].isnull()
    
    # apply intra-day mask
    tt = tt[mask0]
        
    # get same optimal nfft used by pna functions
    nfft_opt, events_required = scorr.get_nfft(nfft, tt.groupby('date')['r1'])
    maxlag = int(min(nfft_opt/2, events_required))
    
    res['maxlag'] = maxlag
    
    # correlations and responses
    # ------------------------------------------------------------------------
    kwargs = {'subtract_mean': False, 'norm': 'cov'}
    if not group:
        # one continuous time-series
        s  = tt['sign'].values
        c  = tt['change'].values
        r  = tt['r1'].values
        sc = tt['sc'].values
        sn = tt['sn'].values
        # "normal" correlations
        res['sacorr'] = scorr.fftcrop(scorr.acorr(s, **kwargs), maxlag)
        res['cccorr'] = scorr.fftcrop(scorr.xcorr(sc, sc, **kwargs), maxlag)
        res['nncorr'] = scorr.fftcrop(scorr.xcorr(sn, sn, **kwargs), maxlag)
        res['cncorr'] = scorr.fftcrop(scorr.xcorr(sc, sn, **kwargs), maxlag)
        res['nccorr'] = scorr.fftcrop(scorr.xcorr(sn, sc, **kwargs), maxlag)
        # triple cross correlations
        if 'hdim2' in models:
            res['ccccorr'] = scorr.x3corr(
                c, sc, sc, nfft=2*maxlag, pad=maxlag, **kwargs
            )
            res['nnccorr'] = scorr.x3corr(
                c, sn, sn, nfft=2*maxlag, pad=maxlag, **kwargs
            )
            res['cnccorr'] = scorr.x3corr(
                c, sc, sn, nfft=2*maxlag, pad=maxlag, **kwargs
            )
        # responses
        signed_lags, S, R   = prop.response(r, s, maxlag=maxlag)
        signed_lags, Sc, Rc = prop.response(r, sc, maxlag=maxlag)
        signed_lags, Sn, Rn = prop.response(r, sn, maxlag=maxlag)
        res['signed_lags']  = signed_lags
        res['S'] = S 
        res['R'] = R
        res['Sc'] = Sc
        res['Rc'] = Rc
        res['Sn'] = Sn
        res['Rn'] = Rn
        res['cmean'] = tt['change'].mean()
    else:
        # treat days separately
        res['sacorr'] = scorr.acorr_grouped_df(
            tt, ['sign'], nfft=nfft, return_df=False, **kwargs)[0][:maxlag]
        res['cccorr'] = scorr.fftcrop(
            scorr.xcorr_grouped_df(
                tt, ['sc','sc'], nfft=nfft, return_df=False, **kwargs
            )[0], 
            maxlag
        )
        res['nncorr'] = scorr.fftcrop(
            scorr.xcorr_grouped_df(
                tt, ['sn', 'sn'], nfft=nfft, return_df=False, **kwargs
            )[0], 
            maxlag
        )
        res['cncorr'] = scorr.fftcrop(
            scorr.xcorr_grouped_df(
                tt, ['sc', 'sn'], nfft=nfft, return_df=False, **kwargs
            )[0], 
            maxlag
        )
        res['nccorr'] = scorr.fftcrop(
            scorr.xcorr_grouped_df(
                tt, ['sn', 'sc'], nfft=nfft, return_df=False, **kwargs
            )[0], 
            maxlag
        )
        # triple cross correlations
        if 'hdim2' in models:
            res['ccccorr'] = scorr.x3corr_grouped_df(
                tt, ['change', 'sc', 'sc'], nfft=nfft, **kwargs
            )[0]
            res['nnccorr'] = scorr.x3corr_grouped_df(
                tt, ['change', 'sn', 'sn'], nfft=nfft, **kwargs
            )[0]
            res['cnccorr'] = scorr.x3corr_grouped_df(
                tt, ['change', 'sc', 'sn'], nfft=nfft, **kwargs
            )[0]
        # responses
        signed_lags, S, R   = prop.response_grouped_df(
            tt, ['r1', 'sign'], nfft=nfft
        )
        signed_lags, Sc, Rc = prop.response_grouped_df(
            tt, ['r1', 'sc'],   nfft=nfft
        )
        signed_lags, Sn, Rn = prop.response_grouped_df(
            tt, ['r1', 'sn'],   nfft=nfft
        )
        res['signed_lags'] = signed_lags
        res['S'] = S
        res['R'] = R
        res['Sc'] = Sc
        res['Rc'] = Rc
        res['Sn'] = Sn
        res['Rn'] = Rn
        res['cmean'] = tt.groupby('date')['change'].mean().mean()
        
    # kernels
    # ------------------------------------------------------------------------

    # TIM1
    if 'tim1' in models:
        res['g'] = prop.calibrate_tim1(res['sacorr'], res['S'], maxlag=maxlag)
    
    # TIM2
    ## estimate kernels
    if 'tim2' in models:
        gn, gc = prop.calibrate_tim2(
            res['nncorr'],
            res['cccorr'],
            res['cncorr'],
            res['nccorr'],
            res['Sn'],
            res['Sc'],
            maxlag=maxlag
        )
        res['gc'] = gc
        res['gn'] = gn

    # HDIM2
    if 'hdim2' in models:
        kn, kc = prop.calibrate_hdim2(
            res['nnccorr'],
            res['ccccorr'],
            res['cnccorr'],
            res['Sn'], 
            res['Sc'], 
            maxlag=maxlag,
        )
        res['kn'] = kn
        res['kc'] = kc
    if 'hdim2_x2' in models:
        kn, kc = prop.calibrate_hdim2(
            prop.corr_mat(res['nncorr'], maxlag=maxlag),
            prop.corr_mat(res['cccorr'], maxlag=maxlag),
            prop.corr_mat(res['cncorr'], maxlag=maxlag),
            res['Sn'], 
            res['Sc'], 
            maxlag=maxlag,
        )
        res['kn_x2'] = kn
        res['kc_x2'] = kc
    return res
    
# Run Models
# ============================================================================

def calc_models(
        dbc,
        nfft              = 'pad', # also for calibration
        group             = False,      # "
        calibrate         = True,
        split_by          = 'sample',
        rshift            = 0, ## sync to causal return, not r1
        models            = ['cs','tim1','tim2','hdim2','hdim2_x2'],
        smooth_kernel     = True
    ):
    """Add propagator(-like) models to trades.  
    
    Pass a dict: calibration is added to dict, not returned.
    Pass trades DataFrame directly: calibration is returned as DataFrame(s)
    
    See also: calibrate_models, aggregate_impact.add_models_to_trades
    """
    # normalise inputs (dict / df)
    if 'tt' in dbc:
        tt = dbc['tt']
    else:
        tt = dbc
        dbc = {'tt': tt}
        
    # kernel preprocessing
    if smooth_kernel:
        if type(smooth_kernel) == dict:
            kern = lambda x: prop.smooth_tail_rbf(x **smooth_kernel)
        else:
            kern = lambda x: prop.smooth_tail_rbf(x)
    else:
        kern = lambda x: x
    
    # propagator ingredients
    complete_data_columns(tt, split_dates=split_by)

    # get masks for different samples (groups of events, normally days)
    masks   = get_trade_split(tt, split_by)
    samples = masks.keys()
    
    for i in range(len(samples)):
        # get calibration for a sample
        ## get a sample name and the corresponding mask
        s = samples[i-1]
        m = masks[s]
        
        ## calculate now or rely on existing calibration?
        if calibrate:
            cal = calibrate_models(
                tt.loc[m], nfft=nfft, group=group, models=models
            )
            if 'cal' in dbc:
                dbc['cal'][s] = cal
            else:
                dbc['cal'] = {s: cal}
        else:
            cal = dbc['cal'][s]
        
        # switch mask to a different sample
        ## keeping the above calibration which is then out-of-sample
        s = samples[i]
        m = masks[s]

        # now run the models
        ## model output -> trades df (sync to regular return, not r1)
        if not group:
            ## treat everything in the sample as one continuous series
            if 'cs' in models:
                tt.loc[m,'r_cs'] = (
                    tt.loc[m,'change'] * tt.loc[m,'sign']
                ).shift(rshift)
                #tt['r_cps'] = (
                #    np.random.permutation(tt['change']) * tt['sign']
                #).shift(1)
            if 'tim1' in models:
                tt.loc[m,'r_tim1']  = shift(
                    prop.tim1(
                        tt.loc[m,'sign'], 
                        kern(cal['g'])
                    ),
                    rshift
                )
            if 'tim2' in models:
                tt.loc[m,'r_tim2'] = shift(
                    prop.tim2(
                       tt.loc[m,'sign'], 
                       tt.loc[m,'change'], 
                       kern(cal['gn']), 
                       kern(cal['gc'])
                   ),
                   rshift
                )
            if 'hdim2' in models:
                tt.loc[m,'r_hdim2'] = shift(
                    prop.hdim2(
                        tt.loc[m,'sign'], 
                        tt.loc[m,'change'], 
                        kern(cal['kn']), 
                        kern(cal['kc'])
                    ),
                    rshift
                )
            if 'hdim2_x2' in models:
                tt.loc[m,'r_hdim2_x2'] = shift(
                    prop.hdim2(
                        tt.loc[m,'sign'], 
                        tt.loc[m,'change'], 
                        kern(cal['kn_x2']), 
                        kern(cal['kc_x2'])
                    ),
                    rshift
                )
        else:
            # simulate models for each day in the sample independently
            g = tt.loc[m].groupby('date')
            rcs       = []
            rtim1     = []
            rtim2     = []
            rhdim2    = []
            rhdim2_x2 = []
            # kernels (smoothing takes time so don't don't in the loop)
            g_tim1   = kern(cal['g'])
            gn_tim2  = kern(cal['gn'])
            gc_tim2  = kern(cal['gc'])
            kn_hdim2 = kern(cal['kn'])
            kc_hdim2 = kern(cal['kc'])
            kn_hdim2_x2 = kern(cal['kn_x2'])
            kc_hdim2_x2 = kern(cal['kc_x2'])
            # simulate for all groups
            for i, (gk, gv) in enumerate(g):
                rcs.append((gv['change'].astype(float) * gv['sign']))
                if 'tim1' in models:
                    rtim1.append(
                        shift(
                            prop.tim1(
                                gv['sign'], g_tim1
                            ),
                            rshift
                        )
                    )
                if 'tim2' in models:
                    rtim2.append(
                        shift(
                             prop.tim2(
                                gv['sign'], 
                                gv['change'], 
                                gn_tim2, 
                                gc_tim2
                            ),
                            rshift
                        )
                    )
                if 'hdim2' in models:
                    rhdim2.append(
                        shift(
                            prop.hdim2(
                                gv['sign'], 
                                gv['change'], 
                                kn_hdim2, 
                                kc_hdim2
                            ),
                            rshift
                        )
                    )
                if 'hdim2_x2' in models:
                    rhdim2_x2.append(
                        shift(
                            prop.hdim2(
                                gv['sign'], 
                                gv['change'], 
                                kn_hdim2_x2, 
                                kc_hdim2_x2
                            ),
                            rshift
                        )
                    )
            if 'cs' in models:     tt.loc[m,'r_cs']    = np.concatenate(rcs)
            if 'tim1' in models:   tt.loc[m,'r_tim1']  = np.concatenate(rtim1)
            if 'tim2' in models:   tt.loc[m,'r_tim2']  = np.concatenate(rtim2)
            if 'hdim2' in models:  tt.loc[m,'r_hdim2'] = np.concatenate(rhdim2)
            if 'hdim2_x2' in models:
                tt.loc[m,'r_hdim2_x2'] = np.concatenate(rhdim2_x2)
        # end if group
        
    # end loop over samples